import csv
import logging
import os
from datetime import datetime
from typing import List, Literal, Optional

from prompeteer.prompt.prompt import Variable
from prompeteer.prompt.prompt_config import PromptConfig
from prompeteer.providers.llm_client import LLMClient
from prompeteer.providers.llm_request import LLMRequest

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

clients = {}


class Result:
    def __init__(self, response_text: str, request_text: Optional[str] = None):
        self.response_text = response_text
        self.request_text = request_text


def run_prompt(prompt_config_file_path: str,
               input_csv: str,
               output_csv: str = None,
               destination: Literal['file', 'console'] = 'console',
               row_numbers_to_process: List[int] = None,
               include_prompt: bool = False):
    results = _run_prompts(input_csv, prompt_config_file_path, row_numbers_to_process, include_prompt)
    if destination == 'console':
        handle_output_to_console(results)
    elif destination == 'file':
        handle_output_to_file(results, input_csv, output_csv)


def _run_prompts(input_csv: str, prompt_config_file_path: str, row_numbers_to_process: List[int],
                 include_prompt: bool) -> List[Result]:
    results: List[Result] = []
    prompt_config: PromptConfig = PromptConfig.from_file(prompt_config_file_path)
    # Open the input CSV file and process each row
    with open(input_csv, mode='r', newline='', encoding='utf-8') as file:
        reader = csv.DictReader(file, delimiter='|', quotechar='"')
        for csv_row_number, variables_row in enumerate(reader):
            if row_numbers_to_process is None or csv_row_number in row_numbers_to_process:
                try:
                    csv_row_variables: List[Variable] = [Variable(name, value) for name, value in
                                                         variables_row.items()]
                    llm_request: LLMRequest = prompt_config.to_llm_request(variables_to_inject=csv_row_variables)
                    llm_client = LLMClient.get_client_by_provider(prompt_config.llm_provider)
                    llm_response_text: str = llm_client.call(llm_request)
                    prompt_text = llm_request.to_text() if include_prompt else None
                    results.append(Result(response_text=llm_response_text, request_text=prompt_text))
                except Exception as e:
                    logger.error(f"Error processing row {csv_row_number}: {e}")
                    continue  # Continue with the next row, logging the error for this one
    return results


def handle_output_to_file(results: List[Result], input_csv: str, output_csv: str = None):
    if output_csv is None:
        # if not provided, create the output csv file in the same location where the input_csv is located
        output_csv = create_output_file_name(os.path.dirname(input_csv), "result")
    with open(output_csv, 'w', newline='') as out_file:
        fieldnames = ['response']
        if any(result.request_text for result in results):
            fieldnames.insert(0, 'request')
        writer = csv.DictWriter(out_file, fieldnames=fieldnames, delimiter=",", quoting=csv.QUOTE_ALL, escapechar='\\')
        writer.writeheader()
        for result in results:
            # mandatory output fields
            out_row = {'response': result.response_text}

            # optional output fields
            if result.request_text:
                out_row['request'] = result.request_text.encode('utf_8').decode('unicode_escape')
            writer.writerow(out_row)


def handle_output_to_console(results: List[Result]):
    for result in results:
        if result.request_text:
            print("Request:", result.request_text)
        print("Response:", result.response_text)
        print("--------------------------------------------------------------------------------------------")


def create_output_file_name(directory, output_prefix) -> str:
    return os.path.join(directory, f"{output_prefix}_{datetime.now().strftime('%Y%m%d%H%M%S')}.csv")



import logging
import os
from typing import Dict, Any

from azure.identity import DefaultAzureCredential, get_bearer_token_provider
from openai import AzureOpenAI
from openai.types.chat import ChatCompletion

from prompeteer.providers.azure_openai.azure_llm_request import AzureLLMRequest
from prompeteer.providers.llm_client import LLMClient

token_provider = get_bearer_token_provider(
    DefaultAzureCredential(), "https://cognitiveservices.azure.com/.default"
)
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class AzureOpenAiClient(LLMClient):
    def __init__(self):
        self.azure_resource_to_client_map: Dict[str, AzureOpenAI] = {}

    def call(self, llm_request: AzureLLMRequest) -> str:
        azure_resource_name = os.getenv('AZURE_OPENAI_RESOURCE_NAME')
        if azure_resource_name is None:
            logger.error("AZURE_OPENAI_RESOURCE_NAME environment variable is not set")
            raise Exception("AZURE_OPENAI_RESOURCE_NAME environment variable is not set")

        client = self.get_client_for_resource(azure_resource_name)

        params: Dict[str, Any] = {
            'messages': [{'role': msg.role, 'content': msg.content} for msg in llm_request.messages],
            'model': llm_request.model
        }

        # Add optional parameters only if they are not None
        if llm_request.temperature is not None:
            params['temperature'] = llm_request.temperature
        if llm_request.top_p is not None:
            params['top_p'] = llm_request.top_p
        if llm_request.n is not None:
            params['n'] = llm_request.n
        if llm_request.stream is not None:
            params['stream'] = llm_request.stream
        if llm_request.stop is not None:
            params['stop'] = llm_request.stop
        if llm_request.max_tokens is not None:
            params['max_tokens'] = llm_request.max_tokens
        if llm_request.presence_penalty is not None:
            params['presence_penalty'] = llm_request.presence_penalty
        if llm_request.frequency_penalty is not None:
            params['frequency_penalty'] = llm_request.frequency_penalty
        if llm_request.logit_bias is not None:
            params['logit_bias'] = llm_request.logit_bias
        if llm_request.user is not None:
            params['user'] = llm_request.user

        response: ChatCompletion = client.chat.completions.create(**params)

        if not response.choices:
            raise Exception('Failed to call Azure OpenAI Chat response')
        return response.choices[0].message.content

    def get_client_for_resource(self, azure_resource_name: str) -> AzureOpenAI:
        if azure_resource_name not in self.azure_resource_to_client_map:
            self.azure_resource_to_client_map[azure_resource_name] = AzureOpenAI(
                api_version="2024-02-01",
                azure_endpoint=f"https://{azure_resource_name}.openai.azure.com",
                azure_ad_token_provider=token_provider
            )
        return self.azure_resource_to_client_map[azure_resource_name]



import logging
from abc import abstractmethod, ABC
from typing import Dict

from prompeteer.prompt.prompt import LLMProvider
from prompeteer.providers.llm_request import LLMRequest

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class LLMClient(ABC):
    clients: Dict[str, 'LLMClient'] = {}

    @abstractmethod
    def call(self, llm_request: LLMRequest) -> str:
        raise NotImplementedError("LLM provider not implemented")

    @classmethod
    def get_client_by_provider(cls, provider: LLMProvider) -> 'LLMClient':
        assert provider is not None, "Provider must be provided"
        if provider.value not in cls.clients:
            if provider == LLMProvider.azure:
                logger.info("Initializing Azure OpenAI LLM Client")
                from prompeteer.providers.azure_openai.azure_openai_client import AzureOpenAiClient
                return AzureOpenAiClient()
            elif provider == LLMProvider.aws:
                logger.info("Initializing AWS Bedrock LLM Client")
                from prompeteer.providers.aws_bedrock.aws_bedrock_client import AwsBedrockClient
                return AwsBedrockClient()
            else:
                logger.error(f"Unknown LLM Provider {provider}")
                raise Exception(f"Unknown LLM Provider {provider}")
        logger.info(f"LLM Client for {provider} initialized successfully")
        return cls.clients[provider.value]



import json
import logging
from typing import Dict, Any

import boto3

from prompeteer.providers.aws_bedrock.aws_llm_request import AWSLLMRequest

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)
from tenacity import retry, stop_after_attempt, wait_exponential, retry_if_exception

from prompeteer.providers.llm_client import LLMClient


class Message:
    def __init__(self, content: str, role: str):
        self.content: str = content
        self.role: str = role


class AwsBedrockClient(LLMClient):
    def __init__(self):
        self.client = boto3.client(service_name='bedrock-runtime', region_name='us-east-1')

    @retry(stop=stop_after_attempt(7), wait=wait_exponential(multiplier=1, min=2, max=10),
           retry=retry_if_exception(lambda e: True))
    def call(self, llm_request: AWSLLMRequest) -> str:
        try:
            params: Dict[str, Any] = {
                'messages': [{'role': msg.role, 'content': msg.content} for msg in llm_request.messages],
                'anthropic_version': 'bedrock-2023-05-31',
                'max_tokens': llm_request.max_tokens
            }

            # Add optional parameters only if they are not None
            if llm_request.temperature is not None:
                params['temperature'] = llm_request.temperature
            if llm_request.top_p is not None:
                params['top_p'] = llm_request.top_p
            if llm_request.top_k is not None:
                params['top_k'] = llm_request.top_k
            if llm_request.stop_sequence is not None:
                params['stop_sequence'] = llm_request.stop_sequence
            if llm_request.system is not None:
                params['system'] = llm_request.system

            response = self.client.invoke_model(body=json.dumps(params), modelId=llm_request.model)
            response_body = json.loads(response.get('body').read())

            if len(response_body['content']) == 0:
                raise ValueError("Empty response content, retrying...")

            return response_body['content'][0]['text']

        except Exception as e:
            logger.error(f"An error occurred: {e}")
            raise e



# Author: Yoaz Menda
import re
from typing import List

from prompeteer.prompt.prompt import DeclaredVariable


def camel_to_snake(name):
    """Convert camelCase to snake_case."""
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def normalize_keys(obj):
    """Recursively convert all keys in the dictionary from camelCase to snake_case."""
    if isinstance(obj, dict):
        new_dict = {}
        for k, v in obj.items():
            new_key = camel_to_snake(k)
            new_dict[new_key] = normalize_keys(v)
        return new_dict
    elif isinstance(obj, list):
        return [normalize_keys(item) for item in obj]
    else:
        return obj


def get_declared_variables(variables: List[dict]) -> List[DeclaredVariable]:
    declared_variables = []
    for info in variables:
        variable = DeclaredVariable(name=info['name'], required=info['required'])
        declared_variables.append(variable)
    return declared_variables



import json
from typing import List, Optional, Union, Dict

from prompeteer.prompt.prompt import DeclaredVariable
from prompeteer.providers.llm_request import LLMRequest


class Message:
    def __init__(self, content: str, role: str):
        self.content: str = content
        self.role: str = role


class AzureLLMRequest(LLMRequest):

    def __init__(self,
                 declared_variables: List[DeclaredVariable],
                 model: str,
                 messages: List[Dict],
                 temperature: Optional[float] = None,
                 top_p: Optional[float] = None,
                 n: Optional[int] = None,
                 stream: Optional[bool] = None,
                 stop: Optional[Union[str, List]] = None,
                 max_tokens: Optional[int] = None,
                 presence_penalty: Optional[float] = None,
                 frequency_penalty: Optional[float] = None,
                 logit_bias: Optional[Dict] = None,
                 user: str = None):
        super().__init__(declared_variables)
        self.declared_variables: List[DeclaredVariable]
        self.model = model
        self.messages: List[Message] = [Message(content=msg['content'], role=msg['role']) for msg in messages]
        self.temperature = temperature
        self.top_p = top_p
        self.n = n
        self.stream = stream
        self.stop = stop
        self.max_tokens = max_tokens
        self.presence_penalty = presence_penalty
        self.frequency_penalty = frequency_penalty
        self.logit_bias = logit_bias
        self.user = user

    def to_text(self) -> str:
        result = [{'content': message.content, 'role': message.role} for message in self.messages]
        return json.dumps(result)

    def get_texts_to_inject(self) -> List[str]:
        return [message.content for message in self.messages]  # only consider the contents for variable injection

    def _set_injected_texts(self, injected_texts: List[str]) -> None:
        for i, text in enumerate(injected_texts):
            self.messages[i].content = text



import logging
import re
from abc import abstractmethod, ABC
from typing import List

from prompeteer.prompt.prompt import Variable, DeclaredVariable

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class LLMRequest(ABC):
    def __init__(self, declared_variables: List[DeclaredVariable]):
        self.declared_variables = declared_variables

    @abstractmethod
    def to_text(self) -> str:
        """
        Converts the llm request prompt text to a presentable output format
        """
        raise NotImplementedError

    @abstractmethod
    def get_texts_to_inject(self) -> List[str]:
        """
        Returns a list of all text parts of the prompt that are candidates for variable substitution
        """
        raise NotImplementedError

    def validate_variables_to_inject(self, variables_to_inject: List[Variable]):
        placeholder_pattern = re.compile(r'\{([^}]+)\}')
        all_placeholders = set()

        # Collect all placeholders from the texts
        for text in self.get_texts_to_inject():
            if text:
                all_placeholders.update(placeholder_pattern.findall(text))

        declared_variable_names = {var.name for var in self.declared_variables}

        # 1. Validate that all declared variables appear in the prompt text as placeholders
        for declared_var in declared_variable_names:
            if declared_var not in all_placeholders:
                logger.error(f"Declared variable '{declared_var}' not found in the prompt text.")
                raise Exception(f"Declared variable '{declared_var}' not found in the prompt text.")

        # 2. Validate that every variable placeholder has a corresponding declared variable
        for placeholder in all_placeholders:
            if placeholder not in declared_variable_names:
                logger.error(f"Placeholder '{placeholder}' does not have a corresponding declared variable.")
                raise Exception(
                    f"Placeholder '{placeholder}' does not have a corresponding declared variable.")

        # 3. Validate required variables
        variables_to_inject_names = {var.name for var in variables_to_inject}
        for declared_var in self.declared_variables:
            if declared_var.required and declared_var.name not in variables_to_inject_names:
                logger.error(f"Required variable '{declared_var.name}' is missing in the variables to inject.")
                raise Exception(
                    f"Required variable '{declared_var.name}' is missing in the variables to inject.")

        # 4. Validate the variables_to_inject list
        for var in variables_to_inject:
            if var.name not in all_placeholders:
                logger.error(f"Variable '{var.name}' in variables_to_inject has no matching placeholder in the text.")
                raise Exception(
                    f"Variable '{var.name}' in variables_to_inject has no matching placeholder in the text.")

    def inject_variables(self, variables_to_inject: List[Variable]) -> None:
        variables_map = {var.name: var.value for var in variables_to_inject}

        def replace_placeholders(text: str) -> str:
            if not text:
                return text

            def replacer(match):
                var_name = match.group(1)
                return variables_map.get(var_name, '') if var_name in variables_map else ''

            return re.sub(r'\{([^}]+)\}', replacer, text)

        texts_to_inject = self.get_texts_to_inject()
        injected_texts = [replace_placeholders(text) for text in texts_to_inject]

        self._set_injected_texts(injected_texts)

    @abstractmethod
    def _set_injected_texts(self, injected_texts: List[str]) -> None:
        raise NotImplementedError



from enum import Enum
from logging import getLogger

logger = getLogger(__name__)


class DeclaredVariable:
    def __init__(self, name: str, required: bool):
        self.name = name
        self.required = required


class Variable:
    def __init__(self, name: str, value: str):
        self.name: str = name
        self.value: str = value


class Settings:
    def __init__(self):
        pass

    def to_dict(self):
        return {k: v for k, v in self.__dict__.items() if v is not None}


class LLMProvider(Enum):
    azure = "azure"
    aws = "aws"


def camel_to_snake(name):
    return ''.join(['_' + i.lower() if i.isupper() else i for i in name]).lstrip('_')



import json
from typing import List, Optional, Dict

from prompeteer.prompt.prompt import DeclaredVariable
from prompeteer.providers.llm_request import LLMRequest


class Message:
    def __init__(self, content: str, role: str):
        self.content: str = content
        self.role: str = role


class AWSLLMRequest(LLMRequest):

    def __init__(self,
                 declared_variables: List[DeclaredVariable],
                 model: str,
                 max_tokens: int,
                 messages: List[Dict],
                 system: Optional[str] = None,
                 temperature: Optional[float] = None,
                 top_p: Optional[float] = None,
                 top_k: Optional[float] = None,
                 stop_sequence: Optional[List[str]] = None):
        super().__init__(declared_variables)
        self.model = model
        self.max_tokens = max_tokens
        self.messages: List[Message] = [Message(content=msg['content'], role=msg['role']) for msg in messages]
        self.system = system
        self.temperature = temperature
        self.top_p = top_p
        self.top_k = top_k
        self.stop_sequence = stop_sequence

    def to_text(self) -> str:
        messages = [{'content': message.content, 'role': message.role} for message in self.messages]
        result = {'messages': messages}
        if self.system is not None:
            result['system'] = self.system
        return json.dumps(result)

    def get_texts_to_inject(self) -> List[str]:
        result = [message.content for message in self.messages]
        result.append(self.system)
        return result

    def _set_injected_texts(self, injected_texts: List[str]) -> None:
        for i, text in enumerate(injected_texts[:-1]):
            self.messages[i].content = text
        if self.system is not None:
            self.system = injected_texts[-1]



from setuptools import setup, find_packages

# Read the contents of your README file
with open("public_README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='prompeteer',
    version='0.3.4',
    author='Yoaz Menda',
    author_email='yoazmenda@gmail.com',
    description='Prompt Development and Evaluation tool',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=['tests', 'tests.*', 'README.md']),  # Include all packages
    py_modules=['prompeteer'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'pyyaml',
        'openai == 1.28.1',
        'azure-identity',
        'boto3',
        'tenacity'
    ],
)



import logging
from typing import List, Dict, Any, Optional

import yaml

from prompeteer.prompt.prompt import LLMProvider, DeclaredVariable, Variable
from prompeteer.providers.aws_bedrock.aws_llm_request import AWSLLMRequest
from prompeteer.providers.azure_openai.azure_llm_request import AzureLLMRequest
from prompeteer.providers.llm_request import LLMRequest
from prompeteer.utils.utils import normalize_keys, get_declared_variables

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class PromptConfig:
    def __init__(self, version: str, name: str, llm_provider: LLMProvider, prompt_config_dict: Dict[str, Any],
                 variables: List[Dict], schema_version: str):
        self.version: str = version
        self.name: str = name
        self.schema_version: str = schema_version
        self.llm_provider: LLMProvider = llm_provider
        self.llm_request: Dict[str, Any] = prompt_config_dict
        self.declared_variables: List[DeclaredVariable] = get_declared_variables(variables)

    def to_llm_request(self, variables_to_inject: Optional[List[Variable]] = None) -> LLMRequest:
        if self.llm_provider == LLMProvider.azure:
            llm_request: LLMRequest = AzureLLMRequest(**self.llm_request,
                                                      declared_variables=self.declared_variables)
        elif self.llm_provider == LLMProvider.aws:
            llm_request: LLMRequest = AWSLLMRequest(**self.llm_request,
                                                    declared_variables=self.declared_variables)
        else:
            logger.error("Unsupported LLM")
            raise ValueError("Unsupported LLM")
        if variables_to_inject is not None:
            logger.debug("starting variable injection")
            llm_request.validate_variables_to_inject(variables_to_inject)
            llm_request.inject_variables(variables_to_inject=variables_to_inject)
            logger.debug("variable injection done")
        return llm_request

    @classmethod
    def from_file(cls, prompt_config_file_path) -> 'PromptConfig':
        with open(prompt_config_file_path, 'r') as file:
            prompt_config = yaml.safe_load(file)
            # Normalize keys to snake_case
            prompt_config = normalize_keys(prompt_config)

            name = prompt_config['name']
            version = prompt_config['version']
            schema_version = prompt_config['schema_version']
            try:
                provider = LLMProvider[prompt_config['provider']]
            except KeyError as e:
                logger.error(f"LLM Provider key missing: {e}")
                raise
            except ValueError as e:
                logger.error(f"LLM Provider not supported: {e}")
                raise
            prompt_config_dict = prompt_config['request']
            variables = prompt_config['variables']
            return cls(name=name,
                       version=version,
                       llm_provider=provider,
                       prompt_config_dict=prompt_config_dict,
                       variables=variables,
                       schema_version=schema_version)
